/* 
 * txrx.c --
 *
 *    Implementation of the 'txrx' command as a loadable extension.
 *
 * Copyright (c) 1996 Cristian Mata. All rigths reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS : @(#) txrx.c 1.1 96/08/13 15:18:52
 */

#include "txrxInt.h"

#if defined(__WIN32__)
#   define WIN32_LEAN_AND_MEAN
#   include <windows.h>
#   undef WIN32_LEAN_AND_MEAN

/*
 * VC++ has an alternate entry point called DllMain, so we need to rename
 * our entry point.
 */

#   if defined(_MSC_VER)
#    define EXPORT(a,b) __declspec(dllexport) a b
#    define DllEntryPoint DllMain
#   else
#    if defined(__BORLANDC__)
#        define EXPORT(a,b) a _export b
#    else
#        define EXPORT(a,b) a b
#    endif
#   endif
#else
#   define EXPORT(a,b) a b
#endif

/*
 * Declarations for functions defined in this file.
 */

static int           TxrxCmd _ANSI_ARGS_((ClientData clientData,
                                        Tcl_Interp *interp, 
                                        int argc, char **argv));

static int           TapeCmd _ANSI_ARGS_((ClientData clientData, 
                                          Tcl_Interp *interp, 
                                          int argc, char **argv));
static void          DeleteTapeCmd _ANSI_ARGS_((ClientData clientData));

EXTERN EXPORT(int,Txrx_Init) _ANSI_ARGS_((Tcl_Interp *interp));
EXTERN EXPORT(int,Txrx_SafeInit) _ANSI_ARGS_((Tcl_Interp *interp));
     
static void          TxrxCmdDelete _ANSI_ARGS_((ClientData clientData));


/*
 *----------------------------------------------------------------------
 *
 * DllEntryPoint --
 *
 *    This wrapper function is used by Windows to invoke the
 *    initialization code for the DLL.  If we are compiling
 *    with Visual C++, this routine will be renamed to DllMain.
 *    routine.
 *
 * Results:
 *    Returns TRUE;
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

#ifdef __WIN32__
BOOL APIENTRY
DllEntryPoint(hInst, reason, reserved)
     HINSTANCE hInst;        /* Library instance handle. */
     DWORD reason;        /* Reason this function is being called. */
     LPVOID reserved;        /* Not used. */
{
    return TRUE;
}
#endif

/*
 *----------------------------------------------------------------------
 *
 * Txrx_Init --
 *
 *    This procedure initializes module TxRx.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

EXPORT(int,Txrx_Init)(interp)
     Tcl_Interp *interp;
{
    TxrxData *txrxData;
    
    /*
     * Allocate a new Txrx data structure for this interpreter.
     */
    txrxData = (TxrxData *)ckalloc(sizeof(TxrxData));
    Tcl_InitHashTable(&txrxData->tapesTable, TCL_STRING_KEYS);
    txrxData->tapeId = 0;
    txrxData->interp = interp;
    txrxData->isSafe = 1;
    /*
     * Setup the 'Txrx' command.
     */
    txrxData->command = Tcl_CreateCommand(interp, TXRX_NAME, TxrxCmd, 
                                          (ClientData) txrxData, 
                                          TxrxCmdDelete);
    return Tcl_PkgProvide(interp, TXRX_NAME, TXRX_VERSION);
}
/*
 *----------------------------------------------------------------------
 *
 * Txrx_SafeInit --
 *
 *    This procedure initializes module TxRx in a safe interpreter.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

EXPORT(int,Txrx_SafeInit)(interp)
     Tcl_Interp *interp;
{
    TxrxData *txrxData;
    
    /*
     * Allocate a new Txrx data structure for this interpreter.
     */
    txrxData = (TxrxData *)ckalloc(sizeof(TxrxData));
    Tcl_InitHashTable(&txrxData->tapesTable, TCL_STRING_KEYS);
    txrxData->tapeId = 0;
    txrxData->interp = interp;
    txrxData->isSafe = 0;
    /*
     * Setup the 'Txrx' command.
     */
    txrxData->command = Tcl_CreateCommand(interp, TXRX_NAME, TxrxCmd, 
                                          (ClientData) txrxData, 
                                          TxrxCmdDelete);
    return Tcl_PkgProvide(interp, TXRX_NAME, TXRX_VERSION);
}

/*
 *----------------------------------------------------------------------
 *
 * TxrxCmdDelete --
 *
 *    Delete all data associated with TXRX_NAME for this interpreter.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    All tapes in the interpreter are deleted.
 *
 *----------------------------------------------------------------------
 */

static void
TxrxCmdDelete(clientData)
     ClientData clientData;               /* Pointer to TxRxData */
{
    TxrxData *txrxData;
    char cmdBuffer[256];
    char *txrxCmd;
    
    txrxData = (TxrxData *)clientData;
    txrxCmd = Tcl_GetCommandName(txrxData->interp, txrxData->command); 
    
    /* Traverse the hash table and delete all diagrams. */
    sprintf(cmdBuffer, 
            "foreach tape [%s names] {rename $tape {}}", txrxCmd);
    Tcl_GlobalEval(txrxData->interp, cmdBuffer);
    
    Tcl_DeleteHashTable(&txrxData->tapesTable);
    ckfree((char *)txrxData);
}

/*
 *----------------------------------------------------------------------
 *
 * TxrxCmd --
 *
 *    This function implements the "txrx" command. See the manual page
 *  for a description.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
TxrxCmd(clientData, interp, argc, argv)
     ClientData clientData;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
    int c, length, i;
    char tapeName[20], *name;
    Tcl_HashEntry *hPtr;
    int newEntry;
    Tcl_HashSearch search;
    Tcl_Channel protocolChannel;
    Txrx_Tape *tapePtr, *oldTapePtr;
    TxrxData *txrxData = (TxrxData *)clientData;
    int newId;
    int isFile;
    char *inputString;              /* Text buffer for protocol.            */
    int n, sz;
    int incrementId;                /* Should the id be incremented at end. */

    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                         " option ?args...?\"", (char *) NULL);
        return TCL_ERROR;
    }
    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'p') && (strncmp(argv[1], "parse", length) == 0)) {
        if (argc < 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                             " parse file ?tape?\"", (char *) NULL);
            return TCL_ERROR;
        }
        /* 
         * Get the name of the file channel containing the 
         * description of the protocol.
         */
        isFile = 0;
        if (argv[2][0] == '@') {
            if (txrxData->isSafe == 0) {
                /*
                 * Not allowed to open files in unsafe interpreters.
                 */
                Tcl_AppendResult(interp, 
                                 "cannot open files in safe interpreter",
                                 (char *)NULL);
                return TCL_ERROR;
            }
            /*
             * Dealing with a file.
             */
            argv[2] ++;
            protocolChannel = Tcl_OpenFileChannel(interp, argv[2], "r", 0);
            if (protocolChannel == NULL) {
                return TCL_ERROR;
            }
            isFile = 1;
            sz = 10000;
            do {
                inputString = ckalloc(sz);
                if (inputString == NULL) {
                    Tcl_AppendResult(interp, "out of memory", (char *)NULL);
                    Tcl_Close(interp, protocolChannel);
                    return TCL_ERROR;
                }
                if ((n = Tcl_Read(protocolChannel, inputString, sz)) < sz) {
                    break;
                }
                if (Tcl_Seek(protocolChannel, 0, 0) == TCL_ERROR) {
                    Tcl_Close(interp, protocolChannel);
                    return TCL_ERROR;
                }
                ckfree((char *)inputString);
                sz += 10000;
            } while (sz < 1000000);
            if (sz > 1000000) {
                Tcl_AppendResult(interp, "internal error", (char *)NULL);
                return TCL_ERROR;
            }
            inputString[n] = '\0';
        } else {
            inputString = argv[2];
        }
        /*
         * Get the name of the tape, if there is a name.
         */
        incrementId = 0;
        if (argc == 3) {
            incrementId = 1;
            newId = txrxData->tapeId;
            sprintf(tapeName, "txrx%d", newId);
            name = tapeName;
        } else if (strcmp(argv[3], argv[0]) == 0) {
            Tcl_AppendResult(interp, "cannot name tape \"", argv[0], 
                             "\"", (char *)NULL);
            return TCL_ERROR;
        } else {
            name = argv[3];
        }
        /*
         * Parse tape.
         */
        tapePtr = Txrx_CreateTape(interp, inputString);
        if (isFile) {
            ckfree(inputString);
            if (Tcl_Close(interp, protocolChannel) == TCL_ERROR) {
                Txrx_DeleteTape(tapePtr);
                return TCL_ERROR;
            }
        }
        if (tapePtr == NULL) {
            Tcl_AppendResult(interp, "cannot parse protocol", (char *) NULL);
            return TCL_ERROR;
        }
        
        /*
         * Check binding between name and existing tape. 
         */
        hPtr = Tcl_CreateHashEntry(&txrxData->tapesTable, name, &newEntry);
        if (newEntry == 0) {
            oldTapePtr = (Txrx_Tape *) Tcl_GetHashValue(hPtr);
            /*    
             * Cleanup previous tape.
             */
            Tcl_DeleteCommand(interp, 
                              Tcl_GetCommandName(interp, oldTapePtr->command));
            /*
             * Recreate hash entry.
             */
            hPtr = Tcl_CreateHashEntry(&txrxData->tapesTable, name, &newEntry);
            if (newEntry == 0) {
                return TCL_ERROR;
            }
        }
        Tcl_SetHashValue(hPtr, tapePtr);
        tapePtr->handlePtr = hPtr;
        /* 
         * Create a command (TapeCmd) to handle user input. 
         * Setup delete callback for when the tape is deleted.
         */
        tapePtr->command = Tcl_CreateCommand(interp, name, TapeCmd, 
                                             (ClientData) tapePtr, 
                                             DeleteTapeCmd);
        if (tapePtr->command == NULL) {
            /*
             * Creating the command failed, delete the hash table entry.
             */
            Tcl_DeleteHashEntry(hPtr);
            Txrx_DeleteTape(tapePtr);
            Tcl_AppendResult(interp, "creating command \"", name,
                             "\" failed\n", (char *)NULL);
            return TCL_ERROR;
        }
        interp->result = Tcl_GetHashKey(&txrxData->tapesTable, hPtr);
        if (incrementId) {
            txrxData->tapeId = newId + 1;
        }
    } else if ((c == 'd') && (strncmp(argv[1], "delete", length) == 0)) {
        for (i = 2; i < argc; i    ++) {
            hPtr = Tcl_FindHashEntry(&txrxData->tapesTable, argv[i]);
            if (hPtr == NULL) {
                Tcl_AppendResult(interp, "tape \"", argv[i],
                                 "\" doesn't exist", (char *) NULL);
                return TCL_ERROR;
            }
            tapePtr = (Txrx_Tape *) Tcl_GetHashValue(hPtr);
            
            /* 
             * Delete the command itself, as a side effect the hash
             * table entry and the data structure will be deleted.
             */
            Tcl_DeleteCommand(interp, 
                              Tcl_GetCommandName(interp, tapePtr->command));
        } /* end for */
    } else if ((c == 'l') && (strncmp(argv[1], "list", length) == 0)) {
        if (argc != 2) {
            Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                             " list\"", (char *) NULL);
            return TCL_ERROR;
        }
        
        /*
         * Traverse the hash table containing tape names and accumulate
         * current command names - possibly different
         * from the hash key as tapes may have been 'rename'd.
         */
        for (hPtr = Tcl_FirstHashEntry(&txrxData->tapesTable, &search);
             hPtr != NULL; hPtr = Tcl_NextHashEntry(&search)) {
            tapePtr = (Txrx_Tape *) Tcl_GetHashValue(hPtr);
            Tcl_AppendElement(interp, 
                              Tcl_GetCommandName(interp, tapePtr->command));
        }
    } else {
        Tcl_AppendResult(interp, "bad option \"", argv[1],
                         "\": must be parse, list or delete", (char *) NULL);
        return TCL_ERROR;
    }
    /*
     * Note that we must be very careful not to use TCL_DYNAMIC when
     * the result is set.
     */
    
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DeleteTapeCmd --
 *
 *    Called when a tape is deleted via a rename.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Memory is freed, including the hash table entry.
 *
 *----------------------------------------------------------------------
 */

static void
DeleteTapeCmd(clientData)
     ClientData clientData;           /* Tape description. */
{
    Txrx_Tape *tapePtr;
    Tcl_HashEntry *hPtr;

    tapePtr = (Txrx_Tape *)clientData;
    hPtr = tapePtr->handlePtr;
    if (hPtr == NULL) {
        /*printf("Handle ptr is NULL!\n");*/
        return;
    }
    /*
     * Clean hash table entry and free data structure.
     */
    Tcl_DeleteHashEntry(hPtr);
    Txrx_DeleteTape(tapePtr);
}

/*
 *----------------------------------------------------------------------
 *
 * TapeCmd --
 *
 *    Command processing user input for a tape.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    See documentation.
 *
 *----------------------------------------------------------------------
 */

static int
TapeCmd(clientData, interp, argc, argv)
     ClientData clientData;    /* Tape description. */
     Tcl_Interp *interp;    /* Current interpreter. */
     int argc;            /* Number of arguments. */
     char **argv;        /* Argument strings. */
{
    Txrx_Tape *tapePtr = (Txrx_Tape *)clientData;
    Tcl_Channel chanPtr;
    int c, length;
    int addr;

    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                         " option ?args...?\"", (char *) NULL);
        return TCL_ERROR;
    }
    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'i') && (strncmp(argv[1], "info", length) == 0)) {
        /*
         * Enumerate the symbol table and the contents of the tape.
         */
        if (argc < 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                             argv[0], " info arg\"", (char *)NULL);
            return TCL_ERROR;
        }
        c = argv[2][0];
        length = strlen(argv[2]);
        if ((c == 's') && (strncmp(argv[2], "symbols", length) == 0)) {
            return Txrx_PrintSymbols(tapePtr);
        } else if (   (c == 'p') 
                   && (strncmp(argv[2], "program", length) == 0)
                   && (length > 3)) {
            return Txrx_PrintTape(interp, tapePtr);
        } else if (   (c == 'p') 
                   && (strncmp(argv[2], "procedure", length) == 0)
                   && (length > 3)) {
            if (argc != 4) {
                Tcl_AppendResult(interp, "wrong # args: should be \"",
                                 argv[0], " info procedure name\"", 
                                 (char *)NULL);
                return TCL_ERROR;
            }
            if (Txrx_FindProc(tapePtr, argv[3], &addr) == TCL_ERROR) {
                return TCL_ERROR;
            }
            if (Txrx_InfoProc(tapePtr, addr) == TCL_ERROR) {
                return TCL_ERROR;
            }
        } else {
            Tcl_AppendResult(interp, "bad option \"", argv[2],
                             "\": must be symbols, program or procedure", 
                             (char *) NULL);
            return TCL_ERROR;
        }
    } else if (   (c == 'b') && (strncmp(argv[1], "bind", length) == 0) 
               && length >= 2) {
        if (argc != 4) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                             argv[0], " bind command procedure\"",
                             (char *)NULL);
            return TCL_ERROR;
        }
        /*
         * Determine the starting address for the RPC procedure.
         */
        if (Txrx_FindProc(tapePtr, argv[3], &addr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        /*
         * Establish a link between the Tcl procedure name 
         * and the protocol procedure address.
         */
    } else if (   (c == 'b') && strncmp(argv[1], "buffer", length) == 0
               && length >= 2) {
        if (argc < 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                             " buffer cmd ?args?\"", (char *)NULL);
            return TCL_ERROR;
        }
        length = strlen(argv[2]);
        c = argv[2][0];
        if ((c == 'i') && strncmp(argv[2], "initialize", length) == 0) {
            if (argc != 5) {
                Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                                 " buffer init cmd channel\"", (char *)NULL);
                return TCL_ERROR;
            }
            chanPtr = Tcl_GetChannel(interp, argv[4], NULL);
            if (chanPtr == NULL) {
                return TCL_ERROR;
            }
            length = strlen(argv[3]);
            c = argv[3][0];
            if ((c == 'w') && strncmp(argv[3], "write", length) == 0) {
                /*
                 * Initialize the buffer for send data.
                 */
                if (Txrx_DataInit(tapePtr, chanPtr, DATA_INCREMENT) == 
                    TCL_ERROR) {
                    return TCL_ERROR;
                }
            } else if ((c == 'r') && strncmp(argv[3], "read", length) == 0) {
                /*
                 * Initialize data buffer to receive.
                 */
                if (Txrx_DataInit(tapePtr, chanPtr, -1) == TCL_ERROR) {
                    return TCL_ERROR;
                }
            } else {
                Tcl_AppendResult(interp, "invalid command: ", 
                                 "must be read or write", (char *)NULL);
                return TCL_ERROR;
            }
        } else if ((c == 'r') && strncmp(argv[2], "reset", length) == 0) {
            if (Txrx_DataReset(tapePtr) == TCL_ERROR) {
                return TCL_ERROR;
            }
        } else if ((c == 'f') && strncmp(argv[2], "flush", length) == 0) {
            if (Txrx_DataFlush(tapePtr) == TCL_ERROR) {
                return TCL_ERROR;
            }
        } else {
            Tcl_AppendResult(interp, "invalid option: \"", argv[2],
                             "\" must be init, flush or reset", (char *)NULL);
            return TCL_ERROR;
        }
    } else if ((c == 'c') && (strncmp(argv[1], "call", length) == 0)) {
        if (argc < 4) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                             argv[0], " call channel procedure ?args?\"",
                             (char *)NULL);
            return TCL_ERROR;
        }
        chanPtr = Tcl_GetChannel(interp, argv[2], NULL);
        if (chanPtr == NULL) {
            return TCL_ERROR;
        }
        if (Txrx_FindProc(tapePtr, argv[3], &addr) == TCL_ERROR) {
            return TCL_ERROR;
        }
    } else if (   (c == 's') 
               && (strncmp(argv[1], "service", length) == 0)
               && (length >= 3)) {
        if (argc != 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                             argv[0], " service channel\"", (char *)NULL);
            return TCL_ERROR;
        }
        chanPtr = Tcl_GetChannel(interp, argv[2], NULL);
        if (chanPtr == NULL) {
            return TCL_ERROR;
        }
    } else if ((c == 'w') && (strncmp(argv[1], "write", length) == 0)
               && (length >= 3)) {
        if (argc != 4) {
            Tcl_AppendResult(interp, "wrong # args: should be \"", 
                             argv[0], " write datatype list\"", (char *)NULL);
            return TCL_ERROR;
        }
        /*
         * Convert a Tcl list into an XDR buffer data structure. 
         */
        if (Txrx_Write(tapePtr, argv[2], argv[3]) == TCL_ERROR) {
            return TCL_ERROR;
        }
    } else if ((c == 'r') && (strncmp(argv[1], "read", length) == 0)) {
        if (argc != 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"", 
                             argv[0], " read datatype\"", (char *)NULL);
            return TCL_ERROR;
        }
        /*
         * Read a data structure from the network and store it
         * in the interpreter result as a TCL list.
         */
        if (Txrx_Read(tapePtr, argv[2]) == TCL_ERROR) {
            return TCL_ERROR;
        }
    } else {
        Tcl_AppendResult(interp, "bad option \"", argv[1],
                         "\": must be info, bind, buffer, call, ",
                         "read, service or write", 
                         (char *) NULL);
        return TCL_ERROR;
    }
    
    return TCL_OK;
}    
