/*
 * txrx.h --
 *
 *        Declarations of functions and data structures used by Txrx.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) txrx.h 1.1 96/08/13 15:18:53
 */

#ifndef _TXRX
#define _TXRX

#include "tcl.h"

/*
 * Two identifiers are the same if the first MAXIDENTLEN characters
 * are the same. 
 */
#define MAXIDENTLEN           64

/*
 * Enumeration indicating what sort of data transfer is occuring.
 */
typedef enum {NONE, SEND, RECEIVE} DataOp;

/*
 * Data structure used to store a lexical token.
 */
typedef struct token {
    int type;                      /* Token types are defined as constants. */
    char identifier[MAXIDENTLEN+1];/* Identifiers are strings of at most
                                    * MAXIDENTLEN characters. 
                                    */
    double value;                   /* Numeric constant value.               */
    int length;                     /* Number of characters stored in the
                                     * 'identifier' array.
                                     */
} Token;

/*
 * Data structure representing a protocol description.
 */
typedef struct {
    Tcl_HashEntry *handlePtr;       /* Handle to hash table entry.           */
    Tcl_Command command;            /* Command for tape (default "tapeN").   */
    Tcl_Interp *interp;             /* Interpreter for the tape.             */

    char *inText;                   /* Input text while parsing.             */
    int inPos;                      /* Position in input.                    */
    int inLine;                     /* Line and character of input.          */
    int inChar;
    int inputAvailable;             /* Ungetted char available.              */
    int inputChar;                  /* Ungetted character.                   */
    int lexemeAvailable;            /* Token read available.                 */
    Token lexeme;                   /* Token already scanned.                */
    
    Tcl_HashTable constTypeTable;   /* Constants and type names.             */
    
    long *codePtr;                  /* Pointer to code.                      */
    int codePC;                     /* Code assembly counter.                */
    int codeSize;                   /* Size of code.                         */
    int endCode;                    /* Index of next highest byte code.      */
    int entryPoint;                 /* Location where a code module starts.  */
    Tcl_HashTable patchTable;       /* Locations w. type names as values.    */

    int endian;                     /* If non-zero indicates little endian.  */

    long dataValue;                 /* Value of last integral value sent.   
                                     * Used by discriminated union processing.
                                     */
    DataOp dataOp;                  /* Current operation on data buffer.     */
    long *dataPtr;                  /* Buffer area.                          */
    int dataSize;                   /* Size of the data buffer.              */
    int dataPos;                    /* Next available position in buffer.    */
    Tcl_Channel chanPtr;            /* Channel where the data is read/written
                                     * from.
                                     */
    int dataCount;                  /* Number of bytes in the current record
                                     * fragment.
                                     */
    int dataLast;                   /* If 1 indicates last fragment has been
                                     * read.
                                     */
} Txrx_Tape;


/* procedures */
EXTERN Txrx_Tape*          Txrx_CreateTape _ANSI_ARGS_((Tcl_Interp *interp,
                                                        char *input));
EXTERN void                Txrx_DeleteTape _ANSI_ARGS_((Txrx_Tape *tapePtr));

EXTERN int                 Txrx_Write _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                  char *typeNamePtr,
                                                  char *listPtr));
EXTERN int                 Txrx_Read _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                     char *typeNamePtr));
EXTERN int                 Txrx_DataInit _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                      Tcl_Channel chanPtr,
                                                      int size));
EXTERN int                 Txrx_DataFlush _ANSI_ARGS_((Txrx_Tape *tapePtr));
EXTERN int                 Txrx_DataReset _ANSI_ARGS_((Txrx_Tape *tapePtr));

#endif /* _TXRX */
